<?php

interface Obfuscator {
	
	public function encodeNode($value);
	
	public function decodeNode(\PhpParser\Node $node);
	
}

class Obfuscator_rand implements Obfuscator {
	public static $pairs = [
		['str_rot13', 'str_rot13'],
		['convert_uuencode', 'convert_uudecode'],
		['base64_encode', 'base64_decode'],
		['strrev', 'strrev']
	];

	public function encodeNode($value) {
		//$mode = array_rand(self::$pairs); // nondeterministic
		$mode = ((crc32($value) % count(self::$pairs)) + count(self::$pairs) ) % count(self::$pairs);
		
		$ret = new PhpParser\Node\Expr\FuncCall(
			new PhpParser\Node\Name( self::$pairs[$mode][1] ), 
			[
				new PhpParser\Node\Arg(
					new PhpParser\Node\Scalar\String( call_user_func(self::$pairs[$mode][0], $value) )
				)
			]
		);
		$ret->setAttribute('phppack', true);
		$ret->setAttribute('phppack.mode', $mode);
		return $ret;
	}
	
	public function decodeNode(\PhpParser\Node $node) {
		$mode = $node->getAttribute('phppack.mode');
		$raw = call_user_func( self::$pairs[$mode][1], $node->args[0]->value->value );
		return new PhpParser\Node\Scalar\String( $raw );
	}
}

class Obfuscator_rot13 implements Obfuscator {
	public function encodeNode($value) {
		$ret = new PhpParser\Node\Expr\FuncCall(
			new PhpParser\Node\Name("str_rot13"), 
			[
				new PhpParser\Node\Arg(
					new PhpParser\Node\Scalar\String(str_rot13($value))
				)
			]
		);
		$ret->setAttribute('phppack', true);
		return $ret;
	}
	
	public function decodeNode(\PhpParser\Node $node) {
		$raw = str_rot13( $node->args[0]->value->value );
		return new PhpParser\Node\Scalar\String( $raw );
	}
}

class Obfuscator_uuenc implements Obfuscator {
	public function encodeNode($value) {
		$ret = new PhpParser\Node\Expr\FuncCall(
			new PhpParser\Node\Name("convert_uudecode"), 
			[
				new PhpParser\Node\Arg(
					new PhpParser\Node\Scalar\String(convert_uuencode($value))
				)
			]
		);
		$ret->setAttribute('phppack', true);
		return $ret;
	}
	
	public function decodeNode(\PhpParser\Node $node) {
		$raw = convert_uudecode( $node->args[0]->value->value );
		return new PhpParser\Node\Scalar\String( $raw );
	}
}

class Obfuscator_none implements Obfuscator {
	public function encodeNode($value) {
		return new PhpParser\Node\Scalar\String( $value );
	}
	
	public function decodeNode(\PhpParser\Node $node) {
		return $node;
	}
}
