<?php

// phppack.php

error_reporting(E_ALL);
ini_set('display_errors', true);
set_error_handler(function($errno, $errstr, $errfile, $errline) {
    throw new ErrorException($errstr, 0, $errno, $errfile, $errline);
});

// DON'T load PHP-Parser until AFTER the standard declared functions, classes etc have been listed
require( __DIR__ . '/StandardWhitelist.php' );

global $WHITELIST_NAMES;
$WHITELIST_NAMES = array_flip( getWhitelistedNames() );

require( __DIR__ . '/vendor/nikic/PHP-Parser/lib/bootstrap.php' );
require( __DIR__ . '/Obfuscator.php' );

// 

function name_short() {
	global $WHITELIST_NAMES;

	static $next = 'aa';
	
	$ret = $next++;
	
	// Might hit reserved words like 'do' - if so, continue to next
	// Also builtin functions like 'dl'
	
	if (array_key_exists($ret, $WHITELIST_NAMES)) {
		return name_short(); // recurse
	} else {
		return $ret;
	}
};

function name_nilt() {
	static $next = 0;
	return str_replace(
		['0', '1', '2', '3'],
		['I', 'l', 'T', 'N'],
		''.base_convert(++$next, 10, 4)
	);
};

function name_dbqp() {
	static $next = 0;
	return str_replace(
		['0', '1', '2', '3'],
		['d', 'b', 'q', 'p'],
		''.base_convert(++$next, 10, 4)
	);
}

global $NAME_PROCESSOR;
$NAME_PROCESSOR = 'name_short'; // default

//


global $USED_NAMES;
$USED_NAMES = [];

function filterVarName($str) {
	global $NAME_PROCESSOR;
	global $WHITELIST_NAMES;
	global $USED_NAMES;
	
	// Don't ever filter one-character strings
	if (strlen($str) < 2) {
		return $str;
	}
	
	if (array_key_exists($str, $WHITELIST_NAMES)) {
		return $str;
	}
	
	if (! array_key_exists($str, $USED_NAMES)) {
		$USED_NAMES[$str] = $NAME_PROCESSOR();
	}
	
	return $USED_NAMES[$str];
}

function is_name_used($name) {
	global $WHITELIST_NAMES;
	global $USED_NAMES;
	return array_key_exists($name, $WHITELIST_NAMES) || array_key_exists($name, $USED_NAMES);
}

function dump_tokens_to_file($filename) {
	global $USED_NAMES;
	$fh = fopen($filename, 'wb');
	if (! $fh) {
		fwrite( 2 /* STDERR */ , "WARNING: FAILED to open file '$filename' for writing\n" );
	} else {
		global $USED_NAMES;
		foreach( $USED_NAMES as $k => $v ) {
			fwrite($fh, $k."\n");
		}
		fclose($fh);
	}
}

function whitelist_token($token) {
	global $WHITELIST_NAMES;
	$WHITELIST_NAMES[$token] = true;
}

function whitelist_token_multi(array $tokens) {
	global $WHITELIST_NAMES;
	foreach($tokens as $token) {
		$WHITELIST_NAMES[$token] = true;
	}
}

function whitelist_tokens_from_file($filename) {
	whitelist_token_multi( array_filter( explode("\n", file_get_contents($filename)) , 'strlen') );
}

function whitelist_tokens_below_path($whitelist_path, $file_list) {
	$rp = str_replace('\\', '/', realpath($whitelist_path));

	foreach($file_list as $file) {
		$fp = str_replace('\\', '/', realpath($file));
		if (strpos($fp, $rp) === 0) { // found
		
			// break into parts beyond whitelist path
			$remainder = substr($fp, strlen($rp));
		
			// strip file extension
			$raw = substr($remainder, 0, strrpos($remainder, '.'));
			
			// filter
			$tokens = array_filter( explode('/', $raw), 'strlen' );
			
			// save
			whitelist_token_multi( $tokens );
		}
	}
}

// 

class MyPrinter extends PhpParser\PrettyPrinter\Standard {
	protected function pComments(array $comments) {
		return '';
	}
	protected function pStmts(array $nodes, $indent = true) {
		return parent::pStmts($nodes, false); // never indent
	}
	public function pExpr_Array(PhpParser\Node\Expr\Array_ $node) {
        return '[' . $this->pCommaSeparated($node->items) . ']'; // 5.4-style arrays
    }	
    public function pScalar_String(PhpParser\Node\Scalar\String $node) {
        return safestring($node->value); // var_export($node->value, true); // fix bad escaping
    }
	
}

//

class FixupBadObfuscation extends PhpParser\NodeVisitorAbstract {

	public $static_context = false;
	public $obfuscator;

	// PHP 5.4 doesn't allow static $foo = function_call() but we've
	//    already applied this as we can't look up node parents.
	// Check for invalid obfuscation and undo it
	
	// We actually need to traverse the entire below chain again and
	//    remove any of these static calls.
	
	public function checkIsStatic(PhpParser\Node $node) {
		return (
		
			$node instanceof PhpParser\Node\Stmt\StaticVar ||
			(
				$node instanceof PhpParser\Node\Stmt\Property /*&&
				$node->isStatic()*/
			)
		
		);
	}
	
	public function enterNode(PhpParser\Node $node) {
		if ($this->checkIsStatic($node)) {
			$this->static_context = true;
		}
	}
	
	public function leaveNode(PhpParser\Node $node) {
	
		if ($this->checkIsStatic($node)) {
			$this->static_context = false; // done		
		
		} else if ($this->static_context) {
			if ($node->hasAttribute('phppack')) {
				return $this->obfuscator->decodeNode($node);
			}
		}
		
	}

}

//

function node_basicCheck(PhpParser\Node $node) {
	return (
		(
			$node instanceof PhpParser\Node\Expr\Variable ||
			$node instanceof PhpParser\Node\Stmt\StaticVar ||
			$node instanceof PhpParser\Node\Param ||		
			$node instanceof PhpParser\Node\Stmt\Class_ ||
			$node instanceof PhpParser\Node\Stmt\Interface_ || 
			$node instanceof PhpParser\Node\Expr\PropertyFetch ||
			$node instanceof PhpParser\Node\Stmt\PropertyProperty ||
			$node instanceof PhpParser\Node\Expr\StaticPropertyFetch || 
			$node instanceof PhpParser\Node\Stmt\Function_ ||
			$node instanceof PhpParser\Node\Stmt\ClassMethod ||
			$node instanceof PhpParser\Node\Expr\MethodCall ||
			$node instanceof PhpParser\Node\Expr\StaticCall // static method call
			
		) && (
			is_string($node->name)
			// Can happen with PropertyFetch if the property to fetch is a variable variable
		)
	);
}

class FirstPassCollectTokens extends PhpParser\NodeVisitorAbstract {
	public function leaveNode(PhpParser\Node $node) {
		
		if ( $node instanceof PhpParser\Node\Expr\ClosureUse ) {
			filterVarName($node->var);
			
		} elseif (node_basicCheck($node)) {
			filterVarName($node->name);
			
		} elseif ($node instanceof PhpParser\Node\Name) {
			for($i = 0, $e = count($node->parts); $i !== $e; ++$i) {
				filterVarName($node->parts[ $i ]);
			}
		}
		
	}
}

class MyNodeVisitor extends PhpParser\NodeVisitorAbstract {

	public $obfuscator;

	public function enterNode(PhpParser\Node $node) {
		if ($node instanceof PhpParser\Node\Expr\ConstFetch) {
			// do nothing and don't process children
		}
	
	}
	
	public function leaveNode(PhpParser\Node $node) {
		
		// String constants
		
		if ($node instanceof PhpParser\Node\Scalar\String) {
		
			// If the string exactly matches the name of a known function/
			//  class/etc, it's (probably) being used as a callable
			
			// FIXME: Also check whether it's a fully-qualified or partially-
			//  qualified name (e.g. Foo defined but called as "Namespace\Foo"
			//  string)
			
			if (is_name_used($node->value)) {
				$node->value = filterVarName($node->value);
			
			} elseif (! strlen($node->value) || ! preg_match('~[a-zA-Z]~', $node->value)) {
				// don't obfuscate string constant, do nothing
			
			} else {		
				$ret = $this->obfuscator->encodeNode($node->value);
				return $ret; // replace node
			}
		}
		
		// Convert inline-HTML into echo(string) calls
		
		if ($node instanceof PhpParser\Node\Stmt\InlineHTML) {
			$ret = new PhpParser\Node\Stmt\Echo_([
				$this->obfuscator->encodeNode($node->value)
			]);
			return $ret; // replace node
		}
		
		// Regular targets for obfuscation
		// FIXME: Use separate namespaces?
		
		if (
			$node instanceof PhpParser\Node\Expr\ClosureUse
		) {
			$node->var = filterVarName($node->var);
		}
		
		if (node_basicCheck($node)) {
			
			/*
			if (
				$node->name == "call_user_func" ||
				$node->name == "call_user_func_array"
				// also array_map, array_filter, actually anything that takes a callable as argument
			) {
				die("FATAL: call_user_func not yet supported\n");
			}
			*/
			
			$node->name = filterVarName($node->name);
			
		} elseif ($node instanceof PhpParser\Node\Name) {
			for($i = 0, $e = count($node->parts); $i !== $e; ++$i) {
				$node->parts[ $i ] = filterVarName($node->parts[ $i ]);
			}
		}
		
		
		// Return
	}
}

function preprocess($original) {
	$parser     = new PhpParser\Parser(new PhpParser\Lexer);
	$firstPass  = new PhpParser\NodeTraverser;
	
	$checker = new FirstPassCollectTokens;
	$firstPass->addVisitor($checker);
	$stmts = $parser->parse($original);
	
	if (count($stmts) == 1 && $stmts[0] instanceof PhpParser\Node\Stmt\InlineHTML) {
		// No PHP content included - leave as-is
	} else {
		// filter
		$firstPass->traverse($stmts);
	}
}

function obfuscate_ex($original, \Obfuscator $obfuscator) {
	$parser        = new PhpParser\Parser(new PhpParser\Lexer);
	$prettyPrinter = new MyPrinter;

	$secondPass = new PhpParser\NodeTraverser;
	
	$visitor = new MyNodeVisitor;
	$visitor->obfuscator = $obfuscator;	
	$secondPass->addVisitor($visitor);
	
	$fixer = new FixupBadObfuscation;
	$fixer->obfuscator = $obfuscator;
	$secondPass->addVisitor($fixer);

	// parse
	$stmts = $parser->parse($original);

	if (count($stmts) == 1 && $stmts[0] instanceof PhpParser\Node\Stmt\InlineHTML) {
		// No PHP content included - leave as-is
	} else {
		// filter
		$stmts = $secondPass->traverse( $stmts );
	}
	
	// pretty print
	$code = $prettyPrinter->prettyPrintFile($stmts);

	// posthoc filters
	$code = preg_replace('~\n([{}])~', '\\1', $code);
	
	return $code;
}

function safe_get_contents($path) {
	if (file_exists($path)) {
		
		$ret = file_get_contents($path);
		if ($ret === false) {
			throw new Exception("Failed to load file \"". $path . "\"");
		}
		return $ret;
		
	} else if (strlen($path) && $path[0] !== '/' && $path[1] !== ':' && substr($path, 0, 2) !== '\\\\') {
		return safe_get_contents( __DIR__ . DIRECTORY_SEPARATOR . $path );
		
	} else {
		throw new Exception("File not found \"". $path . "\"");
	
	}
}

function safestring($str) {
	$ret = '"';
	for ($i = 0, $e = strlen($str); $i !== $e; ++$i) {
		if ($str[$i] == '\\' || $str[$i] == '"' || $str[$i] == '$') {
			$ret .= '\\'.$str[$i];
		
		} elseif ($str[$i] == '{' && $i + 1 !== $e && $str[$i + 1] == '$') {
			$ret .= '{\\$';
			$i++;
		
		} elseif (ctype_print($str[$i]) || $str[$i] == "\r" || $str[$i] == "\n" || $str[$i] == "\t") {
			$ret .= $str[$i];
			
		} else {
			$ret .= '\\x'.sprintf("%02x", ord($str[$i]));
		}
	}
	return $ret.'"';
}

function packfiles(array $filelist, $should_obfuscate, $should_pack, \Obfuscator $obfuscator, $compress, $handler) {
	
	if ($should_pack) { 
		
		echo "<?php\n\nclass PackedLoader__".$handler." {\n";
		
		echo <<<'EOF'
	protected $position;
	protected $currentfile;
	protected $base_includedir;
	
	public static $FILECACHE;
	
	public function __construct() {
		$this->base_includedir = get_include_path();
	}
	
	public function stream_open($path, $mode, $options, &$opened_path) {
		$path = self::transform_path($path);
		
		if (array_key_exists($path, self::$FILECACHE)) {
			$this->currentfile = $path;
			$this->position = 0;
			
			$dirpath = substr($path, 0, strrpos($path, '/')); // .'/';
			set_include_path(
EOF;
		var_export($handler);
		echo <<<'EOF'
.'://'.$dirpath.PATH_SEPARATOR.$this->base_includedir);
			
			return true;
		} else {
			return false;
		}
	}
	
	public function stream_close() {
		// FIXME: needs support for nested relative include paths
		// set_include_path($this->base_includedir);
	}
	
	public function stream_read($count) {
		$ret = substr(self::$FILECACHE[$this->currentfile], $this->position, $count);
		$this->position += strlen($ret); // might be smaller than count
		return $ret;
	}
	
	public function stream_write($data) {
		$left = substr(self::$FILECACHE[$this->currentfile], 0, $this->position);
		$right = substr(self::$FILECACHE[$this->currentfile], $this->position + strlen($data)); // overwrite
		
		self::$FILECACHE[$this->currentfile] = $left . $data . $right;
		$this->position += strlen($data);
		
		return strlen($data); // bytes written
	}
	
	public function stream_tell() {
		return $this->position;
	}
	
	public function stream_eof() {
		return $this->position >= strlen(self::$FILECACHE[$this->currentfile]);
	}
	
	protected static function transform_path($path) {
		return str_replace('\\', '/', substr($path, strlen(
EOF;
		var_export($handler);
		echo <<<'EOF'
.'://')));
	}
	
	protected static function stat_helper($path) {	
		if (! array_key_exists($path, self::$FILECACHE)) {
			return false; // nonexistent
		}
		
		$mtime = filemtime(__FILE__);
		
		$ret = [
			'dev' => 2,
			'ino' => 0,
			'mode' => 0100777, // octal
			'nlink' => 1,
			'uid' => 1,
			'gid' => 1,
			'rdev' => 2,
			'size' => strlen(self::$FILECACHE[$path]),
			'atime' => $mtime,
			'mtime' => $mtime,
			'ctime' => $mtime,
			'blksize' => 512,
			'blocks' => ceil(strlen(self::$FILECACHE[$path]) / 512)
		];
		
		return array_merge($ret, array_values($ret));
	}
	
	public function stream_stat() {
		return self::stat_helper($this->currentfile);
	}
	
	public function url_stat($path, $flags) {
		return self::stat_helper(self::transform_path($path));
	}
	
	public function stream_seek($offset, $whence) {
		switch ($whence) {
			case SEEK_SET: {
				if ($offset < strlen(self::$FILECACHE[$this->currentfile]) && $offset >= 0) {
					$this->position = $offset;
					return true;
				}
			} break;

			case SEEK_CUR: {
				if ($offset >= 0) {
					$this->position += $offset;
					return true;
				}
			} break;

			case SEEK_END: {
				if (strlen(self::$FILECACHE[$this->currentfile]) + $offset >= 0) {
					$this->position = strlen(self::$FILECACHE[$this->currentfile]) + $offset;
					return true;
				}
			} break;

			default:
				
		}
		return false;
	}
	
}
EOF;
		echo "PackedLoader__".$handler."::\$FILECACHE = [\n";
		
	} // $should_pack
		
	if ($should_obfuscate) {
		
		// First pass
		foreach($filelist as $filename) {
			preprocess( safe_get_contents($filename) );
		}
	}
	
	foreach($filelist as $filename) {
	
		if ($should_pack) {

			$key = '_/' . str_replace('\\', '/', $filename);

			echo (new MyPrinter)->prettyPrintExpr( $obfuscator->encodeNode($key) );

			echo " => \n";

		}
		
		// {
		
		$body = safe_get_contents($filename);
		
		if ($should_obfuscate) {
			$body = obfuscate_ex($body, $obfuscator);
		}
		
		if ($compress) {
			echo "gzuncompress(";
			var_export(gzcompress($body, 9));
			echo ")";
		} else {
			
			if ($should_pack) {
				var_export($body);
			} else {
				echo $body;
			}
		}
		
		// }
		
		if ($should_pack) {		
			echo "\n , \n";
		}
	}

	if ($should_pack) {		
		echo "];\n";
		echo "if (false===stream_wrapper_register(".var_export($handler, true).",".var_export("PackedLoader__".$handler, true).")){\n";
		echo "die(\"FATAL: Failed to load environment\");\n";
		echo "}\n";
		echo 'require '.var_export($handler.'://_/'.str_replace('\\', '/', $filelist[0]), true).';';
	}
}

function usage() {
	echo <<<'EOF'
Usage:
  phppack [options] [--] [file1.php [...]]

Options:
  --compress                      Compress packed output
  --dump-tokens {output.txt}      Write all detected tokens to file
  --filelist {list.txt}           Process every file in list.txt
  --handler {name}                Set internal name for handler and class
  --mangler=short|nilt|dbqp       Select mode for generating nonsense words
  --no-obfuscate                  Do not obfuscate packed files
  --no-pack                       Obfuscate one file only, do not pack.
  --strings=rand|rot13|uuenc|none Select mode for string constants in file
  --whitelist {token}             Don't obfuscate single token
  --whitelist-tokens {input.txt}  Don't obfuscate all tokens in file
  --whitelist-autoload {path}     Don't obfuscate filenames found within path

EOF;

	die(1);
}

function main($args) {
	
	global $NAME_PROCESSOR;
	
	$obfuscator = new Obfuscator_rot13;
	$should_obfuscate = true;
	$mode_is_pack = true;
	$compress = false;
	$handler = 'phppack';
	
	$dump_tokens = null;
	
	$file_list = [];
	$whitelist_paths = [];
	
	$remainder_are_files = false;
	
	for($i = 0, $e = count($args); $i !== $e; ++$i) {
	
		if ($remainder_are_files) {
			$file_list[] = $args[$i];
			continue;
		}
		
		//
	
		if ($args[$i] == "--no-obfuscate") {
			$should_obfuscate = false;
		
		} elseif ($args[$i] == "--no-pack") {
			$mode_is_pack = false;
		
		} elseif ($args[$i] == "--filelist") {
			$file_list = array_merge(
				$file_list,
				explode("\n", safe_get_contents($args[++$i]))
			);
			
		} elseif ($args[$i] == "--handler") {
			$handler = $args[++$i];
			
		} elseif ($args[$i] == '--mangler=short') {
			$NAME_PROCESSOR = 'name_short';
			
		} elseif ($args[$i] == '--mangler=nilt') {
			$NAME_PROCESSOR = 'name_nilt';
			
		} elseif ($args[$i] == '--mangler=dbqp') {
			$NAME_PROCESSOR = 'name_dbqp';
			
		} elseif ($args[$i] == '--strings=rand') {
			$obfuscator = new Obfuscator_rand;
			
		} elseif ($args[$i] == '--strings=rot13') {
			$obfuscator = new Obfuscator_rot13;
			
		} elseif ($args[$i] == '--strings=uuenc') {
			$obfuscator = new Obfuscator_uuenc;
			
		} elseif ($args[$i] == '--strings=none') {
			$obfuscator = new Obfuscator_none;
			
		} elseif ($args[$i] == '--dump-tokens') {
			$dump_tokens = $args[++$i];
			
		} elseif ($args[$i] == '--whitelist') {
			whitelist_token( $args[++$i] );
			
		} elseif ($args[$i] == '--whitelist-tokens') {
			whitelist_tokens_from_file($args[++$i]);
			
		} elseif ($args[$i] == '--whitelist-autoload') {
			$whitelist_paths[] = $args[++$i];
			
		} elseif ($args[$i] == '--') {
			$remainder_are_files = true;
			
		} elseif ($args[$i] == '--compress') {
			$compress = true;
			
		} elseif ($args[$i] == "--help") {
			usage();
			
		} elseif ($args[$i][0] == "-") {
			echo "FATAL: Unrecognised option '". $args[$i]. "'\n";
			usage();
			
		} else {
			$file_list[] = $args[$i];
			
		}
		
	}
	
	if ($compress && !$mode_is_pack) {
		echo "FATAL: --compress cannot be used with --no-pack\n";
		usage();
	}
	
	//
	
	$file_list = array_filter($file_list, 'strlen');	
	if (! count($file_list)) {
		echo "FATAL: No files selected\n";
		usage();
	}
	if (!$mode_is_pack && count($file_list) !== 1) {
		echo "FATAL: Multiple files specified with --no-pack\n";
		usage();
	}
	
	// Process $whitelist_paths
	foreach($whitelist_paths as $whitelist_path) {
		whitelist_tokens_below_path($whitelist_path, $file_list);
	}
	
	// Process
	packfiles( $file_list, $should_obfuscate, $mode_is_pack, $obfuscator, $compress, $handler );
	
	// dump_tokens
	
	if (! is_null($dump_tokens)) {
		dump_tokens_to_file($dump_tokens);
	}
	
	// done
	return 0;
}

return main(array_slice($_SERVER['argv'], 1));
