#!/bin/bash

PHP=/cygdrive/c/bin/php54/php.exe
TESTNUM=0

obfuscate() {
	STASH=$(mktemp)
	cat > $STASH
	STASH_W=$(cygpath -w "$STASH")
	$PHP phppack.php --no-pack "$STASH_W"
	rm $STASH
}

check() {
	INPUT=$( obfuscate )
	EXPECTED=$1
	RESULT=$( echo "$INPUT" | $PHP -n )
	TESTNUM=$(( $TESTNUM + 1 ))
	
	echo "$INPUT" > ./tmp-tests/$(date +%s)-obfus.php
	
	printf "%3d) " "$TESTNUM"
	
	if [[ "$RESULT" == "$EXPECTED" ]] ; then
		echo "Test passed"
	else
		echo "TEST FAILED"
		
		echo "$INPUT" >&2
		echo "Got '$RESULT' expected '$EXPECTED'" >&2
	fi
}

## TEST

check "[3,6,9]" <<'EOF'
<?php
function q($a) {
	return array_map(function($y) { return $y * 3; }, $a);
}
echo json_encode( call_user_func('q', [1,2,3]) );
EOF

## TEST

check "[2,4,6]" <<'EOF' 
<?php

define('MUL_BY', 2);

function x($a) {
	return $a * MUL_BY;
}

$in = [1, 2, 3];
$out = array_map( 'x', $in );

echo json_encode($out);
EOF

## TEST

check 'quux' <<'EOF'
<?php

error_reporting(E_ALL);
ini_set('display_errors', true);

$foo = 'bar';
$x = function() use (&$use) {
	return 'quux';
};
echo $x();
EOF

## TEST

check 'baz' <<'EOF'
<?php

error_reporting(E_ALL);
ini_set('display_errors', true);

class x {
	static $foo = ['bar' => 'baz'];
	static function quux($a) {
		return self::$foo[$a];
	}
}
echo x::quux('bar');
EOF

######

#$PHP phppack.php --no-pack "vendor/nikic/PHP-Parser/lib/PhpParser/Autoloader.php" > ./tmp-autoloader.php
#$PHP phppack.php --no-pack "vendor/nikic/PHP-Parser/lib/PhpParser/PrettyPrinterAbstract.php" > ./tmp-autoloader.php
#$PHP ./tmp-autoloader.php

read -p "Press any key to continue..."
