#!/bin/bash
# phppack has some dependencies, let's try to pack them into itself

PHP=/cygdrive/c/bin/php54/php.exe
mkdir -p tmp-tests
PREFIX=./tmp-tests/$(date +%s)

#

die() {
	echo "$1" >&2
	exit 1
}

hashfile() {
	md5sum "$1" | cut -d' ' -f1
}

requirematch() {
	if [[ $(hashfile "$1") == $(hashfile "$2") ]] ; then
		echo "$3 hash success [" $(hashfile "$1") "] !"
	else
		die "Hash mismatch processing $3"
	fi
}

shouldmatch() {
	if [[ $(hashfile "$1") == $(hashfile "$2") ]] ; then
		echo "$3 hash success [" $(hashfile "$1") "] !"
	else
		echo "WARNING: Hash mismatch processing $3 , continuing"
	fi
}

requireoptions() {
	local RET=$( $PHP "$1" --help || true )
	
	if ! ( echo "$RET" | grep -q "Options:" ) ; then
		die "Couldn't find Options: when running $2, likely failure"
		
	elif ( echo "$RET" | grep -qi "FATAL" ) ; then
		echo "$RET" | grep -i "FATAL"
		die "Got a FATAL when running $2"
		
	else
		echo "$2 load success"
	fi
}

#


# Collect file list

FILELIST=$PREFIX-filelist.txt

echo 'phppack.php' > $FILELIST
echo 'Obfuscator.php' >> $FILELIST
echo 'StandardWhitelist.php' >> $FILELIST
find vendor | grep php$ | grep -v '/test/' >> $FILELIST


# G1: Pack self, no obfuscation

OUTPUT=$PREFIX-packself.php
$PHP phppack.php --no-obfuscate --filelist $(cygpath -w $FILELIST) > $OUTPUT
requireoptions $OUTPUT "Generation 1"


# G2: Pack self, no obfuscation, using G1

OUTPUT2=$PREFIX-packself-g2.php
$PHP $OUTPUT --no-obfuscate --filelist $(cygpath -w $FILELIST) > $OUTPUT2
requirematch $OUTPUT $OUTPUT2 "Generation 2"


# Collect tokens for vendor libraries only

TOKENFILE=$PREFIX-tokens.txt
VONLYFILE=$PREFIX-vonly.txt
tail -n +2 "$FILELIST" > $VONLYFILE

echo "Tokenising vendor libraries..." 

( $PHP phppack.php --filelist $(cygpath -w $VONLYFILE) --dump-tokens $(cygpath -w $TOKENFILE) >/dev/null ) || die "Error listing tokens"


# G3: Pack self, obfuscation of self-only, no string processing, using G2

OUTPUT3=$PREFIX-packself-g3.php
$PHP $OUTPUT2 --strings=none --filelist $(cygpath -w $FILELIST) --whitelist-tokens $(cygpath -w $TOKENFILE) > $OUTPUT3
requireoptions $OUTPUT3 "Generation 3"


# G3A: Pack self, obfuscation of self only, no string processing, using original

OUTPUT3A=$PREFIX-packself-g3a.php
$PHP phppack.php --strings=none --filelist $(cygpath -w $FILELIST) --whitelist-tokens $(cygpath -w $TOKENFILE) > $OUTPUT3A
shouldmatch $OUTPUT3 $OUTPUT3A "Generation 3a"


# G4: Pack self, obfuscation of self only, using G3

OUTPUT4=$PREFIX-packself-g4.php
$PHP $OUTPUT3 --filelist $(cygpath -w $FILELIST) --whitelist-tokens $(cygpath -w $TOKENFILE) > $OUTPUT4
requireoptions $OUTPUT3 "Generation 4"


# G4a: Pack self, obfuscation of self only, using original

OUTPUT4A=$PREFIX-packself-g4a.php
$PHP phppack.php --filelist $(cygpath -w $FILELIST) --whitelist-tokens $(cygpath -w $TOKENFILE) > $OUTPUT4A
shouldmatch $OUTPUT4 $OUTPUT4A "Generation 4a"


# Minimal whitelists for some dynamic dispatch used within PHP-Parser

EXTRAWHITELIST=$PREFIX-whitelist.txt
grep -Eo 'reduceRule[0-9]+' vendor/nikic/PHP-Parser/lib/PhpParser/Parser.php > $EXTRAWHITELIST
grep -Eo 'function [a-zA-Z_]+' vendor/nikic/PHP-Parser/lib/PhpParser/PrettyPrinter/Standard.php | sed 's/function //' >> $EXTRAWHITELIST
grep -Eo 'function [a-zA-Z_]+' vendor/nikic/PHP-Parser/lib/PhpParser/PrettyPrinter/Standard.php | sed 's/function p//' >> $EXTRAWHITELIST
echo p >> $EXTRAWHITELIST
WHITELIST="--whitelist-autoload vendor/nikic/PHP-Parser/lib/ --whitelist-tokens $(cygpath -w "$EXTRAWHITELIST")"


# G5: Full, using G4

OUTPUT5=$PREFIX-packself-g5.php
$PHP $OUTPUT4 --filelist $(cygpath -w $FILELIST) $WHITELIST > $OUTPUT5
requireoptions $OUTPUT4 "Generation 5"


# G5a: Full, using original

OUTPUT5A=$PREFIX-packself-g5a.php
$PHP phppack.php --filelist $(cygpath -w $FILELIST) $WHITELIST > $OUTPUT5A
shouldmatch $OUTPUT5 $OUTPUT5A "Generation 5a"


# G6: Full, using G5

OUTPUT6=$PREFIX-packself-g6.php
$PHP $OUTPUT5 --filelist $(cygpath -w $FILELIST) $WHITELIST > $OUTPUT6
requirematch $OUTPUT5 $OUTPUT6 "Generation 6"


# 
